#include "OgreOcclusionCullingSceneManager.h"
#include "OgreVisibilityOptionsManager.h"
#include <OgreMath.h>
#include <OgreIteratorWrappers.h>
#include <OgreRenderSystem.h>
#include <OgreCamera.h>
#include <OgreLogManager.h>
#include <OgreStringConverter.h>
#include <OgreEntity.h>
#include <OgreSubEntity.h>
#include <OgreMaterialManager.h>
#include <OgreIteratorWrappers.h>
#include <OgreHeightmapTerrainPageSource.h>
#include "VspBspTree.h"
#include "Containers.h"
#include "ViewCellsManager.h"
#include <OgreConfigFile.h>
#include "OgreTypeConverter.h"
#include "OgreMeshInstance.h"
#include "common.h"
#include "OgreBoundingBoxConverter.h"
#include <OgreManualObject.h>


namespace Ogre {

//-----------------------------------------------------------------------
OcclusionCullingSceneManager::OcclusionCullingSceneManager(const String& name, 
							GtpVisibility::VisibilityManager *visManager):
TerrainSceneManager(name),
mVisibilityManager(visManager), 
mShowVisualization(false),
mRenderNodesForViz(false),
mRenderNodesContentForViz(false),
mVisualizeCulledNodes(false),
mLeavePassesInQueue(0),
mDelayRenderTransparents(true),
mUseDepthPass(false),
mIsDepthPassPhase(false),
mUseItemBuffer(false),
mIsItemBufferPhase(false),
mCurrentEntityId(1),
mEnableDepthWrite(true),
mSkipTransparents(false),
mRenderTransparentsForItemBuffer(true),
mExecuteVertexProgramForAllPasses(false),
mIsHierarchicalCulling(false),
mViewCellsLoaded(false),
mUseViewCells(false),
mUseVisibilityFilter(false),
mCurrentViewCell(NULL),
mElementaryViewCell(NULL),
mDeleteQueueAfterRendering(true),
mNormalExecution(false)
{
	Ogre::LogManager::getSingleton().logMessage("creating occlusion culling scene manager");

	mHierarchyInterface = new OctreeHierarchyInterface(this, mDestRenderSystem);
	
	if (0)
	{
		mDisplayNodes = true;
		mShowBoundingBoxes = true;
		mShowBoxes = true;
	}

	// TODO: set maxdepth to reasonable value
	mMaxDepth = 50;
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::InitDepthPass()
{
	MaterialPtr depthMat = MaterialManager::getSingleton().getByName("Visibility/DepthPass");

	if (depthMat.isNull())
    {
		depthMat = MaterialManager::getSingleton().create(
                "Visibility/DepthPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

        mDepthPass = depthMat->getTechnique(0)->getPass(0);
		mDepthPass->setColourWriteEnabled(false);
		mDepthPass->setDepthWriteEnabled(true);
		mDepthPass->setLightingEnabled(false);
	}
	else
	{
		mDepthPass = depthMat->getTechnique(0)->getPass(0);
	}
}
//-----------------------------------------------------------------------
OcclusionCullingSceneManager::~OcclusionCullingSceneManager()
{
	OGRE_DELETE(mHierarchyInterface);
	CLEAR_CONTAINER(mObjects);
	OGRE_DELETE(mCurrentViewCell);
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::InitItemBufferPass()
{
	MaterialPtr itemBufferMat = MaterialManager::getSingleton().
		getByName("Visibility/ItemBufferPass");

	if (itemBufferMat.isNull())
    {
		// Init
		itemBufferMat = MaterialManager::getSingleton().create("Visibility/ItemBufferPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

		mItemBufferPass = itemBufferMat->getTechnique(0)->getPass(0);
		mItemBufferPass->setColourWriteEnabled(true);
		mItemBufferPass->setDepthWriteEnabled(true);
		mItemBufferPass->setLightingEnabled(true);
		//mItemBufferPass->setLightingEnabled(false);
	}
	else
	{
		mItemBufferPass = itemBufferMat->getTechnique(0)->getPass(0);
	}
	//mItemBufferPass->setAmbient(1, 1, 0);
}
//-------------------------------------------------------------------------
#if 1
void OcclusionCullingSceneManager::setWorldGeometry( DataStreamPtr& stream, const String& typeName )
{
    // Clear out any existing world resources (if not default)
    if (ResourceGroupManager::getSingleton().getWorldResourceGroupName() != 
        ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME)
    {
        ResourceGroupManager::getSingleton().clearResourceGroup(
            ResourceGroupManager::getSingleton().getWorldResourceGroupName());
    }
	destroyLevelIndexes();
    mTerrainPages.clear();
    // Load the configuration
    loadConfig(stream);
	initLevelIndexes();

    // Resize the octree, allow for 1 page for now
    float max_x = mOptions.scale.x * mOptions.pageSize;
    float max_y = mOptions.scale.y;
    float max_z = mOptions.scale.z * mOptions.pageSize;

	float maxAxis = std::max(max_x, max_y);
	maxAxis = std::max(maxAxis, max_z);
	resize( AxisAlignedBox( 0, 0, 0, maxAxis, maxAxis, maxAxis ) );
    
    setupTerrainMaterial();

    setupTerrainPages();

 }
#endif
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::PrepareVisualization(Camera *cam)
{
	// add player camera for visualization purpose
	try 
	{
		Camera *c;
		if ((c = getCamera("PlayerCam")) != NULL)
		{
			getRenderQueue()->addRenderable(c);
		}   
    }
    catch (...)
    {
        // ignore
    }

	// add bounding boxes of rendered objects
	if (0)
	for (BoxList::iterator it = mBoxes.begin(); it != mBoxes.end(); ++it)
	{
		getRenderQueue()->addRenderable(*it);
	}

	// set old view cell geometry to invisible
    if (mCurrentViewCell && mCurrentViewCell->GetMesh())
	{
		const int id = mCurrentViewCell->GetId();
		mViewCellsGeometry[id]->_updateRenderQueue(getRenderQueue());
	}

    if (mRenderNodesForViz || mRenderNodesContentForViz)
	{
		// HACK: change node material so it is better suited for visualization
		MaterialPtr nodeMat = MaterialManager::getSingleton().getByName("Core/NodeMaterial");
		nodeMat->setAmbient(1, 1, 0);
		nodeMat->setLightingEnabled(true);
		nodeMat->getTechnique(0)->getPass(0)->removeAllTextureUnitStates();

		for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
		{
			if (mRenderNodesForViz)
			{
				// render the visible leaf nodes
				if ((*it)->numAttachedObjects() && 
					!(*it)->numChildren() && 
					((*it)->getAttachedObject(0)->getMovableType() == "Entity") &&
					(*it)->getAttachedObject(0)->isVisible())
				{
					//getRenderQueue()->addRenderable((*it));
					(*it)->_addToRenderQueue(cam, getRenderQueue(), false);
				}

				// add bounding boxes instead of node itself
				//(*it)->_addBoundingBoxToQueue(getRenderQueue());
			}

			// add renderables itself
			if (0 && mRenderNodesContentForViz) 
			{
				(*it)->_addToRenderQueue(cam, getRenderQueue(), false);
			}
		}
	}	
}
//-----------------------------------------------------------------------
const Pass *OcclusionCullingSceneManager::_setPass(const Pass* pass, bool evenIfSuppressed)
{
	if (mNormalExecution)
	{
		return SceneManager::_setPass(pass);
	}

	// TODO: setting vertex program is not efficient
	//Pass *usedPass = ((mIsDepthPassPhase && !pass->hasVertexProgram()) ? mDepthPass : pass); 
	
	// set depth fill pass if we currently do not make an aabb occlusion query
	const bool useDepthPass = 
		(mIsDepthPassPhase && !mHierarchyInterface->IsBoundingBoxQuery());

	const IlluminationRenderStage savedStage = mIlluminationStage; 
	
	// set illumination stage to NONE so no shadow material is used 
	// for depth pass or for occlusion query
	if (mIsDepthPassPhase || mHierarchyInterface->IsBoundingBoxQuery())
	{
		mIlluminationStage = IRS_NONE;
	}
	
	// --- set vertex program of current pass in order to set correct depth
	if (mExecuteVertexProgramForAllPasses && 
		mIsDepthPassPhase && 
		pass->hasVertexProgram())
	{
		// add vertex program of current pass to depth pass
		mDepthPass->setVertexProgram(pass->getVertexProgramName());

		if (mDepthPass->hasVertexProgram())
		{
			const GpuProgramPtr& prg = mDepthPass->getVertexProgram();
			// Load this program if not done already
			if (!prg->isLoaded())
				prg->load();
			// Copy params
			mDepthPass->setVertexProgramParameters(pass->getVertexProgramParameters());
		}
	}
	else if (mDepthPass->hasVertexProgram()) // reset vertex program
	{
		mDepthPass->setVertexProgram("");
	}
	
	const Pass *usedPass = useDepthPass ? mDepthPass : pass;

	// save old depth write: needed for item buffer
	const bool IsDepthWrite = usedPass->getDepthWriteEnabled();

	// global option which enables / disables depth writes
	if (!mEnableDepthWrite)
	{
		//usedPass->setDepthWriteEnabled(false);
	}
	//else if (mIsItemBufferPass) {usedPass = mItemBufferPass;}
	

	//-- set actual pass here

	const Pass *result = SceneManager::_setPass(usedPass);


	// reset depth write
	if (!mEnableDepthWrite)
	{
		//usedPass->setDepthWriteEnabled(IsDepthWrite);
	}

	// reset illumination stage
	mIlluminationStage = savedStage;

	return result;
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::_findVisibleObjects(Camera* cam,
														bool onlyShadowCasters)
{
	///////////
	//-- set visibility according to pvs of current view cell

	UpdatePvs(cam);

	if (mNormalExecution)
	{
		OctreeSceneManager::_findVisibleObjects(cam, onlyShadowCasters);
		return;
	}

	if (mShowVisualization)
    {
		//////////////
		//-- show visible scene nodes and octree bounding boxes from last frame

		PrepareVisualization(cam);
	}
	else 
	{	
		// hierarchical culling interleaves identification 
		// and rendering of objects in _renderVisibibleObjects

		// for the shadow pass we use only standard rendering
		// because shadows have low occlusion snyway
		if (mShadowTechnique == SHADOWTYPE_TEXTURE_MODULATIVE && 
			mIlluminationStage == IRS_RENDER_TO_TEXTURE)
		{
			OctreeSceneManager::_findVisibleObjects(cam, onlyShadowCasters);
		}

		// only shadow casters will be rendered in shadow texture pass
		if (0) mHierarchyInterface->SetOnlyShadowCasters(onlyShadowCasters);
	}
		
	// lists only used for visualization
	mVisible.clear();
	mBoxes.clear();
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::_renderVisibleObjects()
{
	if (mNormalExecution)
	{
		// the standard octree rendering mode
		OctreeSceneManager::_renderVisibleObjects();
		return;
	}

	InitDepthPass(); 	  // create material for depth pass
	InitItemBufferPass(); // create material for item buffer pass

	// save ambient light to reset later
	ColourValue savedAmbient = mAmbientLight;

	////////////////////
	//-- apply standard rendering for some modes 
	//-- (e.g., the visualization mode, the shadow pass)

	if (mShowVisualization ||
	   (mShadowTechnique == SHADOWTYPE_TEXTURE_MODULATIVE && 
	    mIlluminationStage == IRS_RENDER_TO_TEXTURE))
	{	
		IlluminationRenderStage savedStage = mIlluminationStage; 
	
		if (mShowVisualization) 
		{
			// disable illumination stage to prevent rendering shadows
			mIlluminationStage = IRS_NONE;
		}

		// standard rendering for shadow maps because of performance
		TerrainSceneManager::_renderVisibleObjects();

		mIlluminationStage = savedStage;
	}
	else //-- the hierarchical culling algorithm
	{
		// note matt: this is also called in TerrainSceneManager: really necessary?
		mDestRenderSystem -> setLightingEnabled(false);

		if (mUseItemBuffer)
		{
			// don't render backgrounds for item buffer
			clearSpecialCaseRenderQueues();
			getRenderQueue()->clear();
		}

		////////////////////
		//-- hierarchical culling

		// the objects of different layers (e.g., background, scene, 
		// overlay) must be identified and rendered one after another

		// first render all early skies
		clearSpecialCaseRenderQueues();
		addSpecialCaseRenderQueue(RENDER_QUEUE_BACKGROUND);
		addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_EARLY);
		setSpecialCaseRenderQueueMode(SceneManager::SCRQM_INCLUDE);

		TerrainSceneManager::_renderVisibleObjects();

#ifdef GTP_VISIBILITY_MODIFIED_OGRE
		// delete previously rendered content
		_deleteRenderedQueueGroups();
#endif

		///////////////////
		//-- prepare queue for visible objects (i.e., all but overlay and skies late)

		clearSpecialCaseRenderQueues();
		addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_LATE);
		addSpecialCaseRenderQueue(RENDER_QUEUE_OVERLAY);
	
		// exclude this queues from hierarchical rendering
		setSpecialCaseRenderQueueMode(SceneManager::SCRQM_EXCLUDE);

		// set all necessary parameters for 
		// hierarchical visibility culling and rendering
		InitVisibilityCulling(mCameraInProgress);


		/** 
		* the hierarchical culling algorithm
		* for depth pass: we just find objects and update depth buffer
		* for "delayed" rendering: we render some passes afterwards
		* e.g., transparents, because they need front-to-back sorting
		**/
		
		mVisibilityManager->ApplyVisibilityCulling();

		// delete remaining renderables from queue:
		// all which are not in mLeavePassesInQueue)
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
		_deleteRenderedQueueGroups(mLeavePassesInQueue);
#endif

		/////////////
		//-- reset parameters needed for special rendering
		
		mIsDepthPassPhase = false;
		mIsItemBufferPhase = false;
		mSkipTransparents = false;
		mIsHierarchicalCulling = false;
		
		mLeavePassesInQueue = 0;
		
		if (mUseDepthPass) // the shaded geometry is rendered in a second pass
		{
			// add visible nodes found by the visibility culling algorithm
			NodeList::const_iterator it, it_end = mVisible.end();

			//getRenderQueue()->clear();
			for (it = mVisible.begin(); it != it_end; ++ it)
			{
				(*it)->_addToRenderQueue(mCameraInProgress, getRenderQueue(), false);
			}
		}
		/*const bool mVisualizeViewCells = true;
		if (mVisualizeViewCells)
		{
			MovableObjectsMap::const_iterator mit, mit_end = mViewCellsGeometry.end();
			
			for (mit = mViewCellsGeometry.begin(); mit != mit_end; ++ mit)
			{
				(*mit).second->_updateRenderQueue(getRenderQueue());
			}	
		}*/

		/////////////
		//-- now we can render all remaining queue objects
		//-- used for depth pass, transparents, overlay

        clearSpecialCaseRenderQueues();

		TerrainSceneManager::_renderVisibleObjects();
	} // end hierarchical culling
		
	// HACK: set the new render level index, important to avoid cracks
	// in terrain caused by LOD
	TerrainRenderable::NextRenderLevelIndex();
	
	// reset ambient light
	setAmbientLight(savedAmbient);

	// almost same effect as below
	getRenderQueue()->clear(mDeleteQueueAfterRendering);

	if (0)
	{
		if (!mDeleteQueueAfterRendering)
			getRenderQueue()->clear(true); // finally clear render queue
		else
			OGRE_DELETE(mRenderQueue); // HACK: should rather only be cleared ...
	}

	if (0) WriteLog(); // write out stats
}

//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::_updateSceneGraph(Camera* cam)
{
	if (mNormalExecution)
	{
		OctreeSceneManager::_updateSceneGraph(cam);
		return;
	}

	mVisibilityManager->GetCullingManager()->SetHierarchyInterface(mHierarchyInterface);
	mHierarchyInterface->SetRenderSystem(mDestRenderSystem);

	TerrainSceneManager::_updateSceneGraph(cam);
}
//-----------------------------------------------------------------------
bool OcclusionCullingSceneManager::setOption(const String & key, const void * val)
{
	if (key == "UseDepthPass")
	{
		mUseDepthPass = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "PrepareVisualization")
	{
		mShowVisualization = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesForViz")
	{
		mRenderNodesForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesContentForViz")
	{
		mRenderNodesContentForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyBoxEnabled")
	{
		mSkyBoxEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyPlaneEnabled")
	{
		mSkyPlaneEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyDomeEnabled")
	{
		mSkyDomeEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "VisualizeCulledNodes")
	{
		mVisualizeCulledNodes = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "DelayRenderTransparents")
	{
		mDelayRenderTransparents = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "DepthWrite")
	{
		mEnableDepthWrite = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "UseItemBuffer")
	{
		mUseItemBuffer = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "ExecuteVertexProgramForAllPasses")
	{ 
		mExecuteVertexProgramForAllPasses  = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderTransparentsForItemBuffer")
	{ 
		mRenderTransparentsForItemBuffer  = (*static_cast<const bool *>(val));
		return true;
	}
	else if (key == "DeleteRenderQueue")
	{
		mDeleteQueueAfterRendering = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "NodeVizScale")
	{ 
		OctreeNode::setVizScale(*static_cast<const float *>(val));
		return true;
	}
	if (key == "LoadViewCells")
	{ 
		if (!mViewCellsLoaded)
		{
			String filename(static_cast<const char *>(val));
			mViewCellsLoaded = LoadViewCells(filename);	
		}

		return mViewCellsLoaded;
	}
	if (key == "UseViewCells")
	{ 
		if (mViewCellsLoaded)
		{
			mUseViewCells = *static_cast<const bool *>(val);

			// reset view cell
			OGRE_DELETE(mCurrentViewCell);
			
			if (mUseViewCells)
			{
				mCurrentViewCell = mViewCellsManager->GenerateViewCell();
			}

			mElementaryViewCell = NULL;

			// if we use view cells, all objects are set to false initially
			SetObjectsVisible(!mUseViewCells);
		}

		return true;
	}
	if (key == "NormalExecution")
	{ 
		mNormalExecution = *static_cast<const bool *>(val);
		return true;
	}
	if (key == "UseVisibilityFilter")
	{
		mUseVisibilityFilter = *static_cast<const bool *>(val);
		// set null =>recomputation of the pvs
        mElementaryViewCell = NULL;
		return true;
	}

	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		setOption(key, val) || TerrainSceneManager::setOption(key, val);
}
//-----------------------------------------------------------------------
bool OcclusionCullingSceneManager::getOption(const String & key, void *val)
{
	if (key == "NumHierarchyNodes")
	{
		* static_cast<unsigned int *>(val) = (unsigned int)mNumOctants;
		return true;
	}
	if (key == "VisibilityManager")
	{
		* static_cast<GtpVisibility::VisibilityManager **>(val) = 
			(GtpVisibility::VisibilityManager *)mVisibilityManager;
		return true;
	}
	if (key == "HierarchInterface")
	{
		* static_cast<GtpVisibility::HierarchyInterface **>(val) = 
			(GtpVisibility::HierarchyInterface *)mHierarchyInterface;
		return true;
	}

	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOption(key, val) && TerrainSceneManager::getOption(key, val);
}
//-----------------------------------------------------------------------
bool OcclusionCullingSceneManager::getOptionValues(const String & key, 
													StringVector &refValueList)
{
	return TerrainSceneManager::getOptionValues( key, refValueList);
}
//-----------------------------------------------------------------------
bool OcclusionCullingSceneManager::getOptionKeys(StringVector & refKeys)
{
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOptionKeys(refKeys) || TerrainSceneManager::getOptionKeys(refKeys);
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::setVisibilityManager(GtpVisibility::
														 VisibilityManager *visManager)
{
	mVisibilityManager = visManager;
}
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *OcclusionCullingSceneManager::getVisibilityManager( void )
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::WriteLog()
{
	std::stringstream d;

	d << "Depth pass: " << StringConverter::toString(mUseDepthPass) << ", "
	  << "Delay transparents: " << StringConverter::toString(mDelayRenderTransparents) << ", "
	  << "Use optimization: " << StringConverter::toString(mHierarchyInterface->GetTestGeometryForVisibleLeaves()) << ", "
	  << "Algorithm type: " << mVisibilityManager->GetCullingManagerType() << ", "
	  << "Hierarchy nodes: " << mNumOctants << ", " 
	  << "Traversed nodes: " << mHierarchyInterface->GetNumTraversedNodes() << ", "
	  << "Rendered nodes: " << mHierarchyInterface->GetNumRenderedNodes() << ", "
	  << "Query culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumQueryCulledNodes() << ", "
	  << "Frustum culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumFrustumCulledNodes() << ", "
      << "Queries issued: " << mVisibilityManager->GetCullingManager()->GetNumQueriesIssued() << ", "
	  << "Found objects: " << (int)mVisible.size() << "\n";

	LogManager::getSingleton().logMessage(d.str());
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::renderBasicQueueGroupObjects(RenderQueueGroup* pGroup, 
																QueuedRenderableCollection::OrganisationMode om)
{
    // Basic render loop
    // Iterate through priorities
    RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();

    while (groupIt.hasMoreElements())
    {
        RenderPriorityGroup* pPriorityGrp = groupIt.getNext();

        // Sort the queue first
        pPriorityGrp->sort(mCameraInProgress);

        // Do solids
        renderObjects(pPriorityGrp->getSolidsBasic(), om, true);

		// for correct rendering, transparents must be rendered 
		// after hierarchical culling => don't render them now

        if (mNormalExecution || !mSkipTransparents)
		{
			// Do transparents (always descending)
			renderObjects(pPriorityGrp->getTransparents(), 
				QueuedRenderableCollection::OM_SORT_DESCENDING, true);
		}


    } // for each priority
}

//-----------------------------------------------------------------------
bool OcclusionCullingSceneManager::validatePassForRendering(Pass* pass)
{
	if (mNormalExecution)
	{
		return SceneManager::validatePassForRendering(pass);
	}

	// skip all but first pass if we are doing the depth pass
	if ((mIsDepthPassPhase || mIsItemBufferPhase) && (pass->getIndex() > 0))
	{
		return false;
	}
	// all but first pass
	/*else if ((!mIsDepthPassPhase || mIsItemBufferPhase) && (pass->getIndex() != 0))
	{
		return false;
	}*/

	return SceneManager::validatePassForRendering(pass);
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::_renderQueueGroupObjects(RenderQueueGroup* pGroup,
															QueuedRenderableCollection::OrganisationMode om)
{
	if (mNormalExecution || !mIsItemBufferPhase)
	{
		TerrainSceneManager::_renderQueueGroupObjects(pGroup, om);
		return;
	}
#ifdef  ITEM_BUFFER
	//-- item buffer
	//-- item buffer: render objects using false colors

    // Iterate through priorities
    RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();

	while (groupIt.hasMoreElements())
    {
		RenderItemBuffer(groupIt.getNext());
	}
#endif // ITEM_BUFFER
}
#ifdef ITEM_BUFFER
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::RenderItemBuffer(RenderPriorityGroup* pGroup)
{
	// Do solids
	QueuedRenderableCollection solidObjs = pGroup->getSolidsBasic();//msz

	// ----- SOLIDS LOOP -----
	RenderPriorityGroup::SolidRenderablePassMap::const_iterator ipass, ipassend;
	ipassend = solidObjs.end();

	for (ipass = solidObjs.begin(); ipass != ipassend; ++ipass)
	{
		// Fast bypass if this group is now empty
		if (ipass->second->empty()) 
			continue;

		// Render only first pass of renderable as false color
		if (ipass->first->getIndex() > 0)
			continue;

		RenderPriorityGroup::RenderableList* rendList = ipass->second;
		
		RenderPriorityGroup::RenderableList::const_iterator irend, irendend;
		irendend = rendList->end();
			
		for (irend = rendList->begin(); irend != irendend; ++irend)
		{
			if (0)
			{
				std::stringstream d; d << "itembuffer, pass name: " << 
					ipass->first->getParent()->getParent()->getName();

				LogManager::getSingleton().logMessage(d.str());
			}
			
			RenderSingleObjectForItemBuffer(*irend, ipass->first);
		}
	}

	//-- TRANSPARENT LOOP: must be handled differently from solids

	// transparents are treated either as solids or completely discarded
	if (mRenderTransparentsForItemBuffer)
	{
		QueuedRenderableCollection transpObjs = pGroup->getTransparents(); //msz
		RenderPriorityGroup::TransparentRenderablePassList::const_iterator 
			itrans, itransend;

		itransend = transpObjs.end();
		for (itrans = transpObjs.begin(); itrans != itransend; ++itrans)
		{
			// like for solids, render only first pass
			if (itrans->pass->getIndex() == 0)
			{	
				RenderSingleObjectForItemBuffer(itrans->renderable, itrans->pass);
			}
		}
	}
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::RenderSingleObjectForItemBuffer(Renderable *rend, Pass *pass)
{
	static LightList nullLightList;
	
	int col[4];
	
	// -- create color code out of object id
	col[0] = (rend->getId() >> 16) & 255;
	col[1] = (rend->getId() >> 8) & 255;
	col[2] = rend->getId() & 255;
//	col[3] = 255;

	//mDestRenderSystem->setColour(col[0], col[1], col[2], col[3]);
    
	mItemBufferPass->setAmbient(ColourValue(col[0] / 255.0f,
										    col[1] / 255.0f, 
											col[2] / 255.0f, 1));

	// set vertex program of current pass
	if (mExecuteVertexProgramForAllPasses && pass->hasVertexProgram())
	{
		mItemBufferPass->setVertexProgram(pass->getVertexProgramName());

		if (mItemBufferPass->hasVertexProgram())
		{
			const GpuProgramPtr& prg = mItemBufferPass->getVertexProgram();
			// Load this program if not done already
			if (!prg->isLoaded())
				prg->load();
			// Copy params
			mItemBufferPass->setVertexProgramParameters(pass->getVertexProgramParameters());
		}
	}
	else if (mItemBufferPass->hasVertexProgram())
	{
		mItemBufferPass->setVertexProgram("");
	}

	const Pass *usedPass = _setPass(mItemBufferPass); 


	// render a single object, this will set up auto params if required
	renderSingleObject(rend, usedPass, false, &nullLightList);
}
#endif // ITEM_BUFFER
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *OcclusionCullingSceneManager::GetVisibilityManager()
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::InitVisibilityCulling(Camera *cam)
{
	// reset culling manager stats
	mVisibilityManager->GetCullingManager()->InitFrame(mVisualizeCulledNodes);

	// set depth pass flag before rendering
	mIsDepthPassPhase = mUseDepthPass;

	mIsHierarchicalCulling = true; // during hierarchical culling

	// item buffer needs full ambient lighting to use item colors as unique id
	if (mUseItemBuffer) 
	{
		mIsItemBufferPhase = true;
		setAmbientLight(ColourValue(1,1,1,1));
	}


	// set passes which are stored in render queue 
	// for rendering AFTER hierarchical culling, i.e., passes which need 
	// a special rendering order
	
	mLeavePassesInQueue = 0;

	// if we have the depth pass or use an item buffer, no passes are left in the queue
	if (1 && !mUseDepthPass && !mUseItemBuffer)
	{
		if (mShadowTechnique == SHADOWTYPE_STENCIL_ADDITIVE)
		{
			// TODO: remove this pass because it should be processed during hierarchical culling
			//mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_NOSHADOW;

			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_DECAL;
			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_DIFFUSE_SPECULAR;
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;

			// just render ambient passes
			/*** msz: no more IRS_AMBIENT, see OgreSceneManager.h ***/
			// mIlluminationStage = IRS_AMBIENT; 
			//getRenderQueue()->setSplitPassesByLightingType(true);
		}
	
		if (mShadowTechnique == SHADOWTYPE_STENCIL_MODULATIVE)
		{
			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_NOSHADOW;
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;
		}
	
		// transparents should be rendered after hierarchical culling to 
		// provide front-to-back ordering
		if (mDelayRenderTransparents)
		{
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;
		}
	}

	// skip rendering transparents during the hierarchical culling
	// (because they will be rendered afterwards)
	mSkipTransparents = 
		(mIsDepthPassPhase || (mLeavePassesInQueue & RenderPriorityGroup::TRANSPARENT_PASSES));

	// -- initialise interface for rendering traversal of the hierarchy
	mHierarchyInterface->SetHierarchyRoot(mOctree);
	
	// possible two cameras (one for culling, one for rendering)
	mHierarchyInterface->InitTraversal(mCameraInProgress, 
									   mCullCamera ? getCamera("CullCamera") : NULL,
									   mLeavePassesInQueue);
		
}
//-----------------------------------------------------------------------
OctreeHierarchyInterface *OcclusionCullingSceneManager::GetHierarchyInterface()
{
	return mHierarchyInterface;
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::endFrame()
{
	TerrainRenderable::ResetRenderLevelIndex();
}
//-----------------------------------------------------------------------
Entity* OcclusionCullingSceneManager::createEntity(const String& entityName, 
													const String& meshName)
{
	Entity *ent = SceneManager::createEntity(entityName, meshName);

	for (int i = 0; i < (int)ent->getNumSubEntities(); ++i)
	{
		ent->getSubEntity(i)->setId(mCurrentEntityId);
	}

	// increase counter of entity id values
	++ mCurrentEntityId;

	return ent;
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::renderAdditiveStencilShadowedQueueGroupObjects(
	RenderQueueGroup* pGroup, QueuedRenderableCollection::OrganisationMode om)
{
	// only render solid passes during hierarchical culling
	if (mIsHierarchicalCulling)
	{
		RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();
	    LightList lightList;

		while (groupIt.hasMoreElements())
		{
			RenderPriorityGroup* pPriorityGrp = groupIt.getNext();

			// Sort the queue first
			pPriorityGrp->sort(mCameraInProgress);

			// Clear light list
			lightList.clear();

			// Render all the ambient passes first, no light iteration, no lights
			/*** msz: no more IRS_AMBIENT, see OgreSceneManager.h ***/
			// mIlluminationStage = IRS_AMBIENT;

			OctreeSceneManager::renderObjects(pPriorityGrp->getSolidsBasic(), om, false, &lightList);
			// Also render any objects which have receive shadows disabled
			OctreeSceneManager::renderObjects(pPriorityGrp->getSolidsNoShadowReceive(), om, true);
#if 0		
			std::stringstream d; 
			d << " solid size: " << (int)pPriorityGrp->_getSolidPasses().size()
				<< " solid no shadow size: " << (int)pPriorityGrp->_getSolidPassesNoShadow().size()
				<< "difspec size: " << (int)pPriorityGrp->_getSolidPassesDiffuseSpecular().size()
				<< " decal size: " << (int)pPriorityGrp->_getSolidPassesDecal().size();
			LogManager::getSingleton().logMessage(d.str());
#endif
		}
	}
	else // render the rest of the passes
	{
		OctreeSceneManager::renderAdditiveStencilShadowedQueueGroupObjects(pGroup, om);
	}
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManager::renderModulativeStencilShadowedQueueGroupObjects(
	RenderQueueGroup* pGroup, QueuedRenderableCollection::OrganisationMode om)
{
   if (mIsHierarchicalCulling)
   {
	   // Iterate through priorities
	   RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();

	   while (groupIt.hasMoreElements())
	   {
		   RenderPriorityGroup* pPriorityGrp = groupIt.getNext();

		   // Sort the queue first
		   pPriorityGrp->sort(mCameraInProgress);
		   // Do (shadowable) solids
		   OctreeSceneManager::renderObjects(pPriorityGrp->getSolidsBasic(), om, true);
	   }
   }
   else
   {
	   OctreeSceneManager::renderModulativeStencilShadowedQueueGroupObjects(pGroup, om);
   }
}
//-------------------------------------------------------------------------
void OcclusionCullingSceneManager::SetObjectsVisible(const bool visible)
{
	GtpVisibilityPreprocessor::ObjectContainer::iterator it, it_end = mObjects.end();

	for (it = mObjects.begin(); it != it_end; ++ it)
	{
		OgreMeshInstance *omi = static_cast<OgreMeshInstance *>(*it);
		Entity *ent = omi->GetEntity();
		
		ent->setVisible(visible);
	}
}
//-----------------------------------------------------------------------
bool OcclusionCullingSceneManager::LoadViewCells(const String &filename)
{
	// objects are set to invisible initially
	SetObjectsVisible(false);
	
	// converter between view cell ids and Ogre entites 
	GtpVisibilityPreprocessor::IndexedBoundingBoxContainer iboxes;
	OctreeBoundingBoxConverter bconverter(this);

	// load the view cells assigning the found objects to the pvss
	const bool finalizeViewCells = true;
	mViewCellsManager = 
		GtpVisibilityPreprocessor::ViewCellsManager::LoadViewCells(filename, &mObjects, true, &bconverter);

	if (finalizeViewCells)
	{
		CreateViewCellsGeometry();
	}

	return (mViewCellsManager != NULL);
}
//-------------------------------------------------------------------------
void OcclusionCullingSceneManager::ApplyViewCellPvs(GtpVisibilityPreprocessor::ViewCell *vc, 
													const bool load)
{	// NOTE: should not encounter NULL view cell, 
	// rather apply view cell representing unbounded space then
	if (!vc) 
	{	
		// if no there is no view cell, set everything visible
		SetObjectsVisible(true);
		return;
	}
		
	GtpVisibilityPreprocessor::ObjectPvsMap::const_iterator oit,
			oit_end = vc->GetPvs().mEntries.end();

	////////////
	//-- set PVS of view cell to visible

	for (oit = vc->GetPvs().mEntries.begin(); oit != oit_end; ++ oit)
	{
		if (!(*oit).first) continue;

		OgreMeshInstance *omi = dynamic_cast<OgreMeshInstance *>((*oit).first);
		omi->GetEntity()->setVisible(load);
		//GtpVisibilityPreprocessor::Debug << "assigned id " << omi->GetId() << endl;
	}
}
//-------------------------------------------------------------------------
void OcclusionCullingSceneManager::UpdatePvs(Camera *cam)
{
	if (!(mViewCellsLoaded && mUseViewCells))
		return;

	const GtpVisibilityPreprocessor::Vector3 viewPoint = 
		OgreTypeConverter::ConvertFromOgre(cam->getDerivedPosition());

	GtpVisibilityPreprocessor::ViewCell *newElementary = 
		mViewCellsManager->GetViewCell(viewPoint);

	// elementary view cell did not change => apply same pvs
	if (mElementaryViewCell == newElementary)
		return;

	mElementaryViewCell = newElementary;
	

	//////////////
	//-- unload old pvs

	ApplyViewCellPvs(mCurrentViewCell, false);

	// the new view cell
	GtpVisibilityPreprocessor::ViewCell *viewCell;
		
	if (mUseVisibilityFilter)
	{	
		////////////
		//-- compute new filtered cell

		GtpVisibilityPreprocessor::PrVs prvs;
		mViewCellsManager->GetPrVS(viewPoint, prvs, 5);
		viewCell = prvs.mViewCell;
	}
	else
	{
		viewCell = newElementary;
	}

	///////////////
	//-- load new pvs

	ApplyViewCellPvs(viewCell, true);

	// store pvs
	if (viewCell)
	{
		mCurrentViewCell->SetPvs(viewCell->GetPvs());
		mCurrentViewCell->SetMesh(viewCell->GetMesh());
		mCurrentViewCell->SetId(viewCell->GetId());

		// delete merge tree of filtered view cell
		if (mUseVisibilityFilter)
			mViewCellsManager->DeleteLocalMergeTree(viewCell);
	}
}
//-------------------------------------------------------------------------
void OcclusionCullingSceneManager::CreateViewCellsGeometry()
{
	LogManager::getSingleton().logMessage("creating view cells geometry");

	GtpVisibilityPreprocessor::ViewCellContainer viewCells = mViewCellsManager->GetViewCells();

	GtpVisibilityPreprocessor::ViewCellContainer::const_iterator it, it_end = viewCells.end();
	for (it = viewCells.begin(); it != it_end; ++ it)
	{
		GtpVisibilityPreprocessor::ViewCell *viewCell = *it;

		std::stringstream str;
		str << "processing view cell with id : " << viewCell->GetId();
		LogManager::getSingleton().logMessage(str.str());
		ManualObject *manual = OgreTypeConverter::ConvertToOgre(viewCell->GetMesh(), this);
		mViewCellsGeometry[viewCell->GetId()] = manual;

		// attach to scene node
		getRootSceneNode()->createChildSceneNode()->attachObject(manual);
		manual->setQueryFlags(0); // returned by no query
		// initialy set to invisible
		//manual->setVisible(false);
	}
}

#if 0
//-------------------------------------------------------------------------
void OcclusionCullingSceneManager::TestVisible(SceneNode *node)
{
	// first test for scene node, then for octant (part of the hierarchy)
	if (!node->mVisibleChildren)
		node->setVisible(false);

	node->getOctant()->mVisibleChildren --;
}
//-------------------------------------------------------------------------
void OcclusionCullingSceneManager::TestVisible(Octree *octant)
{
	// first test for scene node, then for octant (part of the hierarchy)
	if (!octant->mVisibleChildren)
	{
		octant->setVisible(false);
	}
}

//-------------------------------------------------------------------------
void OcclusionCullingSceneManager::UpdateVisibility(Entity *ent)
{
	if (!ent->isVisible())
	{
		bool visible = TestVisible(ent->getParentNode());
		
		if (!visible)
			visible = TestVisible(octant->getParentNode());

		if (!visible)
			mHierarchyInterface->pullupVisibility();
	}
}
#endif
//-----------------------------------------------------------------------
const String OcclusionCullingSceneManagerFactory::FACTORY_TYPE_NAME = "OcclusionCullingSceneManager";
//-----------------------------------------------------------------------
void OcclusionCullingSceneManagerFactory::initMetaData(void) const
{
	mMetaData.typeName = FACTORY_TYPE_NAME;
	mMetaData.description = "Scene manager organising the scene on the basis of an octree with advanced occlusion culling (TM).";
	mMetaData.sceneTypeMask = 0xFFFF; // support all types
	mMetaData.worldGeometrySupported = false;
}
//-----------------------------------------------------------------------
SceneManager *OcclusionCullingSceneManagerFactory::createInstance(
		const String& instanceName)
{
	OcclusionCullingSceneManager* tsm = new OcclusionCullingSceneManager(instanceName, visManager);
	
	// Create & register default sources (one per manager)
	HeightmapTerrainPageSource* ps = new HeightmapTerrainPageSource();
	mTerrainPageSources.push_back(ps);
	tsm->registerPageSource("Heightmap", ps);

	return tsm;
}
//-----------------------------------------------------------------------
void OcclusionCullingSceneManagerFactory::destroyInstance(SceneManager* instance)
{
	delete instance;
}

} // namespace Ogre
